import asyncio
import random
import time
from typing import Optional, Dict, Callable, Union, List
from concurrent.futures import ThreadPoolExecutor
from .mouse import Mouse
from .connection import SerialTransport
from .errors import MakcuConnectionError
from .enums import MouseButton
from functools import wraps

def maybe_async(func):
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        try:
            loop = asyncio.get_running_loop()
            async def async_wrapper():
                def execute_sync():
                    return func(self, *args, **kwargs)
                executor = getattr(self, '_executor', None)
                return await loop.run_in_executor(executor, execute_sync)
            return async_wrapper()
        except RuntimeError:
            return func(self, *args, **kwargs)
    
    return wrapper

class MakcuController:
    _BUTTON_LOCK_MAP = {
        MouseButton.LEFT: 'lock_left',
        MouseButton.RIGHT: 'lock_right',
        MouseButton.MIDDLE: 'lock_middle',
        MouseButton.MOUSE4: 'lock_side1',
        MouseButton.MOUSE5: 'lock_side2',
    }
    
    def __init__(self, fallback_com_port: str = "", debug: bool = False, 
                 send_init: bool = True, auto_reconnect: bool = True, 
                 override_port: bool = False) -> None:
        self.transport = SerialTransport(
            fallback_com_port, 
            debug=debug, 
            send_init=send_init,
            auto_reconnect=auto_reconnect,
            override_port=override_port
        )
        self.mouse = Mouse(self.transport)
        self._executor = ThreadPoolExecutor(max_workers=1)
        self._connection_callbacks: List[Callable[[bool], None]] = []
        self._connected = False

    def _check_connection(self) -> None:
        if not self._connected:
            raise MakcuConnectionError("Not connected")

    def _notify_connection_change(self, connected: bool) -> None:
        for callback in self._connection_callbacks:
            try:
                callback(connected)
            except Exception:
                pass

    @maybe_async
    def connect(self) -> None:
        self.transport.connect()
        self._connected = True
        self._notify_connection_change(True)

    @maybe_async
    def disconnect(self) -> None:
        self.transport.disconnect()
        self._connected = False
        self._notify_connection_change(False)
        self._executor.shutdown(wait=False)

    @maybe_async
    def is_connected(self) -> bool:
        return self._connected and self.transport.is_connected()

    @maybe_async
    def click(self, button: MouseButton) -> None:
        self._check_connection()
        self.mouse.press(button)
        self.mouse.release(button)

    @maybe_async
    def double_click(self, button: MouseButton) -> None:
        self._check_connection()
        self.mouse.press(button)
        self.mouse.release(button)
        time.sleep(0.001)
        self.mouse.press(button)
        self.mouse.release(button)

    @maybe_async
    def move(self, dx: int, dy: int) -> None:
        self._check_connection()
        self.mouse.move(dx, dy)

    @maybe_async
    def batch_execute(self, actions: List[Callable[[], None]]) -> None:
        """Execute a batch of actions in sequence.
        
        Args:
            actions: List of callable functions to execute in order
            
        Example:
            makcu.batch_execute([
                lambda: makcu.move(50, 0),
                lambda: makcu.click(MouseButton.LEFT),
                lambda: makcu.move(-50, 0),
                lambda: makcu.click(MouseButton.RIGHT)
            ])
        """
        self._check_connection()
        
        for action in actions:
            try:
                action()
            except Exception as e:
                raise RuntimeError(f"Batch execution failed at action: {e}")

    @maybe_async
    def scroll(self, delta: int) -> None:
        self._check_connection()
        self.mouse.scroll(delta)

    @maybe_async
    def press(self, button: MouseButton) -> None:
        self._check_connection()
        self.mouse.press(button)

    @maybe_async
    def release(self, button: MouseButton) -> None:
        self._check_connection()
        self.mouse.release(button)

    @maybe_async
    def move_smooth(self, dx: int, dy: int, segments: int = 10) -> None:
        self._check_connection()
        self.mouse.move_smooth(dx, dy, segments)

    @maybe_async
    def move_bezier(self, dx: int, dy: int, segments: int = 20,
                    ctrl_x: Optional[int] = None, ctrl_y: Optional[int] = None) -> None:
        self._check_connection()
        if ctrl_x is None:
            ctrl_x = dx // 2
        if ctrl_y is None:
            ctrl_y = dy // 2
        self.mouse.move_bezier(dx, dy, segments, ctrl_x, ctrl_y)

    @maybe_async
    def lock(self, target: Union[MouseButton, str]) -> None:
        self._check_connection()
        
        if isinstance(target, MouseButton):
            if target in self._BUTTON_LOCK_MAP:
                getattr(self.mouse, self._BUTTON_LOCK_MAP[target])(True)
            else:
                raise ValueError(f"Unsupported button: {target}")
        elif target.upper() in ['X', 'Y']:
            if target.upper() == 'X':
                self.mouse.lock_x(True)
            else:
                self.mouse.lock_y(True)
        else:
            raise ValueError(f"Invalid lock target: {target}")

    @maybe_async
    def unlock(self, target: Union[MouseButton, str]) -> None:
        self._check_connection()
        
        if isinstance(target, MouseButton):
            if target in self._BUTTON_LOCK_MAP:
                getattr(self.mouse, self._BUTTON_LOCK_MAP[target])(False)
            else:
                raise ValueError(f"Unsupported button: {target}")
        elif target.upper() in ['X', 'Y']:
            if target.upper() == 'X':
                self.mouse.lock_x(False)
            else:
                self.mouse.lock_y(False)
        else:
            raise ValueError(f"Invalid unlock target: {target}")

    @maybe_async
    def lock_left(self, lock: bool) -> None:
        self._check_connection()
        self.mouse.lock_left(lock)

    @maybe_async
    def lock_middle(self, lock: bool) -> None:
        self._check_connection()
        self.mouse.lock_middle(lock)

    @maybe_async
    def lock_right(self, lock: bool) -> None:
        self._check_connection()
        self.mouse.lock_right(lock)

    @maybe_async
    def lock_side1(self, lock: bool) -> None:
        self._check_connection()
        self.mouse.lock_side1(lock)

    @maybe_async
    def lock_side2(self, lock: bool) -> None:
        self._check_connection()
        self.mouse.lock_side2(lock)

    @maybe_async
    def lock_x(self, lock: bool) -> None:
        self._check_connection()
        self.mouse.lock_x(lock)

    @maybe_async
    def lock_y(self, lock: bool) -> None:
        self._check_connection()
        self.mouse.lock_y(lock)

    @maybe_async
    def lock_mouse_x(self, lock: bool) -> None:
        self.lock_x(lock)

    @maybe_async
    def lock_mouse_y(self, lock: bool) -> None:
        self.lock_y(lock)

    @maybe_async
    def is_locked(self, button: MouseButton) -> bool:
        self._check_connection()
        return self.mouse.is_locked(button)

    @maybe_async
    def get_all_lock_states(self) -> Dict[str, bool]:
        self._check_connection()
        return self.mouse.get_all_lock_states()

    @maybe_async
    def spoof_serial(self, serial: str) -> None:
        self._check_connection()
        self.mouse.spoof_serial(serial)

    @maybe_async
    def reset_serial(self) -> None:
        self._check_connection()
        self.mouse.reset_serial()

    @maybe_async
    def get_device_info(self) -> Dict[str, str]:
        self._check_connection()
        return self.mouse.get_device_info()

    @maybe_async
    def get_firmware_version(self) -> str:
        self._check_connection()
        return self.mouse.get_firmware_version()

    @maybe_async
    def get_button_mask(self) -> int:
        self._check_connection()
        return self.transport.get_button_mask()

    @maybe_async
    def get_button_states(self) -> Dict[str, bool]:
        self._check_connection()
        return self.transport.get_button_states()

    @maybe_async
    def is_pressed(self, button: MouseButton) -> bool:
        self._check_connection()
        return self.transport.get_button_states().get(button.name.lower(), False)

    @maybe_async
    def enable_button_monitoring(self, enable: bool = True) -> None:
        self._check_connection()
        self.transport.enable_button_monitoring(enable)

    @maybe_async
    def set_button_callback(self, callback: Optional[Callable[[MouseButton, bool], None]]) -> None:
        self._check_connection()
        self.transport.set_button_callback(callback)

    @maybe_async
    def on_connection_change(self, callback: Callable[[bool], None]) -> None:
        self._connection_callbacks.append(callback)

    @maybe_async
    def remove_connection_callback(self, callback: Callable[[bool], None]) -> None:
        if callback in self._connection_callbacks:
            self._connection_callbacks.remove(callback)

    @maybe_async
    def click_human_like(self, button: MouseButton, count: int = 1,
                        profile: str = "normal", jitter: int = 0) -> None:
        self._check_connection()

        timing_profiles = {
            "normal": (60, 120, 100, 180),
            "fast": (30, 60, 50, 100),
            "slow": (100, 180, 150, 300),
            "variable": (40, 200, 80, 250),
            "gaming": (20, 40, 30, 60),
        }

        if profile not in timing_profiles:
            raise ValueError(f"Invalid profile: {profile}")

        min_down, max_down, min_wait, max_wait = timing_profiles[profile]

        for i in range(count):
            if jitter > 0:
                dx = random.randint(-jitter, jitter)
                dy = random.randint(-jitter, jitter)
                self.mouse.move(dx, dy)

            self.mouse.press(button)
            time.sleep(random.uniform(min_down, max_down) / 1000.0)
            self.mouse.release(button)
            
            if i < count - 1:
                time.sleep(random.uniform(min_wait, max_wait) / 1000.0)

    @maybe_async
    def drag(self, start_x: int, start_y: int, end_x: int, end_y: int,
             button: MouseButton = MouseButton.LEFT, duration: float = 1.0) -> None:
        self._check_connection()
        
        # Move to start position
        self.move(start_x, start_y)
        time.sleep(0.02)
        
        # Press button
        self.press(button)
        time.sleep(0.02)
        
        # Smooth move to end position
        segments = max(10, int(duration * 30))
        self.move_smooth(end_x - start_x, end_y - start_y, segments)
        
        # Release button
        time.sleep(0.02)
        self.release(button)

    # Context managers for both sync and async
    def __enter__(self):
        if not self.is_connected():
            self.connect()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.disconnect()

    async def __aenter__(self):
        if not await self.is_connected():
            await self.connect()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.disconnect()

    # Legacy async methods for backward compatibility
    async def async_connect(self) -> None:
        """Legacy method - use connect() instead"""
        await self.connect()

    async def async_disconnect(self) -> None:
        """Legacy method - use disconnect() instead"""
        await self.disconnect()

    async def async_click(self, button: MouseButton) -> None:
        """Legacy method - use click() instead"""
        await self.click(button)

    async def async_move(self, dx: int, dy: int) -> None:
        """Legacy method - use move() instead"""
        await self.move(dx, dy)

    async def async_scroll(self, delta: int) -> None:
        """Legacy method - use scroll() instead"""
        await self.scroll(delta)

def create_controller(fallback_com_port: str = "", debug: bool = False, 
                     send_init: bool = True, auto_reconnect: bool = True, 
                     override_port: bool = False) -> MakcuController:
    """Create and connect a controller synchronously"""
    makcu = MakcuController(
        fallback_com_port, 
        debug=debug, 
        send_init=send_init,
        auto_reconnect=auto_reconnect,
        override_port=override_port
    )
    makcu.connect()
    return makcu


async def create_async_controller(fallback_com_port: str = "", debug: bool = False,
                                 send_init: bool = True, auto_reconnect: bool = True, 
                                 override_port: bool = False) -> MakcuController:
    """Create and connect a controller asynchronously"""
    makcu = MakcuController(
        fallback_com_port,
        debug=debug,
        send_init=send_init,
        auto_reconnect=auto_reconnect,
        override_port=override_port
    )
    await makcu.connect()
    return makcu


import pytest
import time
from makcu import MakcuController, MouseButton

@pytest.fixture(scope="session")
def makcu():
    ctrl = MakcuController(fallback_com_port="COM1", debug=False)
    return ctrl


from .controller import (
    MakcuController,
    create_controller,
    create_async_controller,
    maybe_async
)
from .enums import MouseButton
from .errors import MakcuConnectionError

# Version info
__version__ = "2.2.1"
__author__ = "SleepyTotem"

# Main exports
__all__ = [
    'MakcuController',
    'MouseButton',
    'MakcuConnectionError',
    'create_controller',
    'create_async_controller',
    'maybe_async'
]


import sys
import os
from pathlib import Path
from typing import List, NoReturn, Optional, Tuple
import pytest
import time
from makcu import create_controller, MakcuConnectionError, MakcuController
import json
import re
import subprocess
import makcu

makcu_version = makcu.__version__

def debug_console():
    controller = create_controller(debug=True)
    transport = controller.transport

    print("🔧 Makcu Debug Console")
    print("Type a raw command (e.g., km.version()) and press Enter.")
    print("Type 'exit' or 'quit' to leave.")

    command_counter = 0

    while True:
        try:
            cmd = input(">>> ").strip()
            if cmd.lower() in {"exit", "quit"}:
                break
            if not cmd:
                continue

            command_counter += 1
            
            response = transport.send_command(cmd, expect_response=True)
            
            if response and response.strip():
                if response.strip() == cmd:
                    print(f"{cmd}")
                else:
                    print(f"{response}")
            else:
                print("(no response)")

        except Exception as e:
            print(f"⚠️ Error: {e}")

    controller.disconnect()
    print("Disconnected.")

def test_port(port: str) -> None:
    try:
        print(f"Trying to connect to {port}...")
        makcu = MakcuController(fallback_com_port=port, send_init=False, override_port=True)
        makcu.connect()
        if makcu.is_connected:
            print(f"✅ Successfully connected to {port}.")
        makcu.disconnect()
    except MakcuConnectionError as e:
        if "FileNotFoundError" in str(e):
            print(f"❌ Port {port} does not exist. Please check the port name.")
        else:
            print(f"❌ Failed to connect to {port}: ") 
    except Exception as e:
        print(f"❌ Unexpected error: {e}")

def check_pytest_html_installed() -> bool:
    """Check if pytest-html is installed."""
    try:
        import pytest_html
        return True
    except ImportError:
        return False

def find_writable_directory() -> Path:
    """Find a writable directory for the HTML report."""
    # Try current working directory first
    cwd = Path.cwd()
    if os.access(cwd, os.W_OK):
        return cwd
    
    # Try user's home directory
    home = Path.home()
    if os.access(home, os.W_OK):
        return home
    
    # Try temp directory as last resort
    import tempfile
    return Path(tempfile.gettempdir())

def parse_html_results(html_file: Path) -> Tuple[List[Tuple[str, str, int]], int]:
    if not html_file.exists():
        raise FileNotFoundError(f"HTML report not found: {html_file}")
   
    with open(html_file, 'r', encoding='utf-8') as f:
        content = f.read()
   
    match = re.search(r'data-jsonblob="([^"]*)"', content)
    if not match:
        raise ValueError("Could not find JSON data in HTML report")
   
    json_str = match.group(1)
    json_str = json_str.replace('&#34;', '"').replace('&amp;#x27;', "'").replace('&amp;', '&')
   
    try:
        data = json.loads(json_str)
    except json.JSONDecodeError as e:
        raise ValueError(f"Failed to parse JSON data: {e}")
   
    test_results = []
    total_ms = 0
   
    skip_tests = {'test_connect_to_port'}
   
    for test_id, test_data_list in data.get('tests', {}).items():
        test_name = test_id.split('::')[-1]
        
        # Skip tests that are in skip_tests
        if test_name in skip_tests:
            continue
           
        for test_data in test_data_list:
            status = test_data.get('result', 'UNKNOWN')
            duration_str = test_data.get('duration', '0 ms')
           
            duration_match = re.search(r'(\d+)\s*ms', duration_str)
            duration_ms = int(duration_match.group(1)) if duration_match else 0
            
            # Always add test to results
            test_results.append((test_name, status, duration_ms))
            
            # Only add time to total if it's not a cleanup test
            if 'cleanup' not in test_name.lower():
                total_ms += duration_ms
   
    return test_results, total_ms

def run_tests() -> NoReturn:
    # Check if pytest-html is installed
    if not check_pytest_html_installed():
        print("❌ pytest-html is not installed. Please install it via:")
        print("   pip install pytest-html")
        sys.exit(1)
    
    try:
        from rich.console import Console
        from rich.table import Table
        from rich.progress import Progress, SpinnerColumn, TextColumn, BarColumn, TimeElapsedColumn
        from rich.panel import Panel
        from rich.align import Align
        from rich import print as rprint
        from rich.text import Text

        console = Console()

        header = Panel.fit(
            f"[bold cyan]Makcu Test Suite {makcu_version}[/bold cyan]\n[dim]High-Performance Python Library[/dim]",
            border_style="bright_blue"
        )
        console.print(Align.center(header))
        console.print()

        package_dir: Path = Path(__file__).resolve().parent
        test_file: Path = package_dir / "test_suite.py"
        
        # Find writable directory and create HTML path
        writable_dir = find_writable_directory()
        html_file: Path = writable_dir / "latest_pytest.html"
        
        # Clean up old report if it exists
        if html_file.exists():
            try:
                html_file.unlink()
            except Exception:
                pass

        console.print(f"[dim]Running pytest to generate: {html_file}[/dim]")
        console.print(f"[dim]Working directory: {Path.cwd()}[/dim]")

        start_time = time.time()

        with Progress(
            SpinnerColumn(),
            TextColumn("[progress.description]{task.description}"),
            BarColumn(),
            TextColumn("[progress.percentage]{task.percentage:>3.0f}%"),
            TimeElapsedColumn(),
            console=console,
            transient=True
        ) as progress:
            task = progress.add_task("[cyan]Running tests...", total=100)

            # Run pytest with explicit output capturing
            result = subprocess.run(
                [
                    sys.executable, "-m", "pytest",
                    str(test_file),
                    "--rootdir", str(package_dir),
                    "-q",
                    "--tb=no",
                    "--html", str(html_file),
                    "--self-contained-html",
                    "-v"  # Add verbose to help debug
                ],
                capture_output=True,
                text=True
            )

            progress.update(task, completed=100)

        # Check if HTML file was created
        if not html_file.exists():
            console.print(f"[red]❌ HTML report was not created at: {html_file}[/red]")
            console.print(f"[yellow]pytest exit code: {result.returncode}[/yellow]")
            if result.stdout:
                console.print("[yellow]stdout:[/yellow]")
                console.print(result.stdout)
            if result.stderr:
                console.print("[red]stderr:[/red]")
                console.print(result.stderr)
            
            # Try to run tests without HTML report
            console.print("\n[yellow]Running tests without HTML report...[/yellow]")
            result2 = subprocess.run(
                [sys.executable, "-m", "pytest", str(test_file), "-v"],
                capture_output=True,
                text=True
            )
            console.print(result2.stdout)
            sys.exit(1)

        try:
            test_results, total_ms = parse_html_results(html_file)
        except (FileNotFoundError, ValueError) as e:
            console.print(f"[red]❌ Failed to parse test results: {e}[/red]")
            console.print(f"[yellow]⚠️ pytest exit code: {result.returncode}[/yellow]")
            sys.exit(1)

        elapsed_time = time.time() - start_time

        table = Table(title="[bold]Test Results[/bold]", show_header=True, header_style="bold magenta")
        table.add_column("Test", style="cyan", no_wrap=True)
        table.add_column("Status", justify="center")
        table.add_column("Time", justify="right", style="yellow")
        table.add_column("Performance", justify="center")

        passed = failed = skipped = 0

        for test_name, status, duration_ms in test_results:
            display_name = test_name.replace("test_", "").replace("_", " ").title()

            if status.upper() == "PASSED":
                if display_name.lower().startswith("cleanup"):
                    status_text = ""
                    passed += 1
                else:
                    status_text = "[green]✅ PASSED[/green]"
                    passed += 1
            elif status.upper() == "FAILED":
                status_text = "[red]❌ FAILED[/red]"
                failed += 1
            elif status.upper() == "SKIPPED":
                status_text = "[yellow]⏭️ SKIPPED[/yellow]"
                skipped += 1
            else:
                status_text = status

            time_str = f"{duration_ms}ms" if duration_ms else "-"
            if duration_ms <= 3:
                perf = "[green]Excellent[/green]"
            elif duration_ms <= 5:
                perf = "[cyan]Great[/cyan]"
            elif duration_ms <= 10:
                perf = "[yellow]Good[/yellow]"
            elif duration_ms > 0:
                if display_name.lower().startswith("cleanup"):
                    perf = ""
                else:
                    perf = "[red]🐌 Needs work[/red]"
            else:
                perf = "-"

            table.add_row(display_name, status_text, time_str, perf)

        console.print("\n")
        console.print(table)
        console.print()

        summary = Table.grid(padding=1)
        summary.add_column(style="bold cyan", justify="right")
        summary.add_column(justify="left")
        summary.add_row("Total Tests:", str(len(test_results)))
        summary.add_row("Passed:", f"[green]{passed}[/green]")
        summary.add_row("Failed:", f"[red]{failed}[/red]" if failed else str(failed))
        summary.add_row("Skipped:", f"[yellow]{skipped}[/yellow]" if skipped else str(skipped))
        summary.add_row("Total Time:", f"{elapsed_time:.2f}s")
        summary.add_row("Avg Time/Test:", f"{total_ms/len(test_results):.1f}ms" if test_results else "0ms")

        console.print(Align.center(Panel(summary, title="[bold]Summary[/bold]", border_style="blue", expand=False)))
        console.print()

        if test_results:
            avg_time = total_ms / len(test_results)
            if avg_time < 3:
                perf_text = Text("Performance: ELITE - Ready for 360Hz+ gaming!", style="bold bright_green")
            elif avg_time < 5:
                perf_text = Text("Performance: EXCELLENT - Ready for 240Hz+ gaming!", style="bold green")
            elif avg_time < 10:
                perf_text = Text("Performance: GREAT - Ready for 144Hz gaming!", style="bold cyan")
            else:
                perf_text = Text("Performance: GOOD - Suitable for standard gaming", style="bold yellow")
        else:
            perf_text = Text("⚠️ No test results parsed. Check your test suite.", style="bold red")

        console.print(Align.center(Panel(perf_text, border_style="green")))
        
        # Print the location of the HTML report
        console.print(f"\n[dim]HTML report saved to: {html_file}[/dim]")
        
        sys.exit(0 if failed == 0 else 1)

    except ImportError:
        print("📦 Rich not installed. Install it via `pip install rich` for enhanced output.")
        print("\nFallback to raw pytest output...\n")

        package_dir: Path = Path(__file__).resolve().parent
        test_file: Path = package_dir / "test_suite.py"
        
        # Find writable directory
        writable_dir = find_writable_directory()
        html_file: Path = writable_dir / "latest_pytest.html"
        
        print(f"HTML report will be saved to: {html_file}")

        # Use subprocess instead of pytest.main for better control
        result = subprocess.run(
            [
                sys.executable, "-m", "pytest",
                str(test_file),
                "--rootdir", str(package_dir),
                "-q",
                "--tb=no",
                "--html", str(html_file),
                "--self-contained-html"
            ],
            capture_output=True,
            text=True
        )

        if not html_file.exists():
            print(f"\n❌ HTML report was not created. pytest exit code: {result.returncode}")
            if result.stdout:
                print("stdout:", result.stdout)
            if result.stderr:
                print("stderr:", result.stderr)
            sys.exit(1)

        try:
            test_results, total_ms = parse_html_results(html_file)
            passed = sum(1 for _, status, _ in test_results if status.upper() == "PASSED")
            failed = sum(1 for _, status, _ in test_results if status.upper() == "FAILED")
            skipped = sum(1 for _, status, _ in test_results if status.upper() == "SKIPPED")
            
            print(f"\n📊 Results: {passed} passed, {failed} failed, {skipped} skipped")
            if test_results:
                avg_time = total_ms / len(test_results)
                print(f"⏱️ Average time per test: {avg_time:.1f}ms")
        except (FileNotFoundError, ValueError):
            print("\n⚠️ Could not parse HTML results for summary")

        if result.returncode != 0:
            print("\n❌ Some tests failed.")
        else:
            print("\n✅ All tests passed.")

        sys.exit(result.returncode)
    
def main() -> None:
    args: List[str] = sys.argv[1:]

    if not args:
        print("Usage:")
        print("  python -m makcu --debug")
        print("  python -m makcu --testPort COM3")
        print("  python -m makcu --runtest")
        return

    if args[0] == "--debug":
        debug_console()
    elif args[0] == "--testPort" and len(args) == 2:
        test_port(args[1])
    elif args[0] == "--runtest":
        run_tests()
    else:
        print(f"Unknown command: {' '.join(args)}")
        print("Usage:")
        print("  python -m makcu --debug")
        print("  python -m makcu --testPort COM3")
        print("  python -m makcu --runtest")

if __name__ == "__main__":
    main()


import serial
import threading
import time
from typing import Optional, Dict, Callable
from serial.tools import list_ports
from dataclasses import dataclass
from collections import deque
from concurrent.futures import Future
import logging
import asyncio
from .errors import MakcuConnectionError, MakcuTimeoutError
from .enums import MouseButton

logger = logging.getLogger(__name__)

@dataclass
class PendingCommand:
    command_id: int
    command: str
    future: Future
    timestamp: float
    expect_response: bool = True
    timeout: float = 0.1

@dataclass
class ParsedResponse:
    command_id: Optional[int]
    content: str
    is_button_data: bool = False
    button_mask: Optional[int] = None

class SerialTransport:
    
    BAUD_CHANGE_COMMAND = bytearray([0xDE, 0xAD, 0x05, 0x00, 0xA5, 0x00, 0x09, 0x3D, 0x00])
    DEFAULT_TIMEOUT = 0.1
    MAX_RECONNECT_ATTEMPTS = 3
    RECONNECT_DELAY = 0.1
    

    BUTTON_MAP = (
        'left', 'right', 'middle', 'mouse4', 'mouse5'
    )
    
    BUTTON_ENUM_MAP = (
        MouseButton.LEFT,
        MouseButton.RIGHT,
        MouseButton.MIDDLE,
        MouseButton.MOUSE4,
        MouseButton.MOUSE5,
    )

    def __init__(self, fallback: str = "", debug: bool = False, 
                 send_init: bool = True, auto_reconnect: bool = True, 
                 override_port: bool = False) -> None:

        self._fallback_com_port = fallback
        self.debug = debug
        self.send_init = send_init
        self.auto_reconnect = auto_reconnect
        self.override_port = override_port
        
        if not hasattr(SerialTransport, '_thread_counter'):
            SerialTransport._thread_counter = 0
            SerialTransport._thread_map = {}

        # Log version info during initialization
        try:
            from makcu import __version__
            version = __version__
            self._log(f"Makcu version: {version}")
        except ImportError:
            self._log("Makcu version info not available")
        
        self._log(f"Initializing SerialTransport with params: fallback='{fallback}', debug={debug}, send_init={send_init}, auto_reconnect={auto_reconnect}, override_port={override_port}")

        self._is_connected = False
        self._reconnect_attempts = 0
        self.port: Optional[str] = None
        self.baudrate = 115200
        self.serial: Optional[serial.Serial] = None
        self._current_baud: Optional[int] = None
        

        self._command_counter = 0
        self._pending_commands: Dict[int, PendingCommand] = {}
        self._command_lock = threading.Lock()
        

        self._parse_buffer = bytearray(1024)
        self._buffer_pos = 0
        self._response_queue = deque(maxlen=100)
        

        self._button_callback: Optional[Callable[[MouseButton, bool], None]] = None
        self._last_button_mask = 0
        self._button_states = 0
        

        self._stop_event = threading.Event()
        self._listener_thread: Optional[threading.Thread] = None
                

        self._ascii_decode_table = bytes(range(128))
        
        self._log("SerialTransport initialization completed")


    def _log(self, message: str, level: str = "INFO") -> None:
        if not self.debug:
            return
            
        timestamp = time.strftime("%H:%M:%S", time.localtime())
        thread_id = threading.get_ident()
        
        # Map thread ID to a simple number
        if thread_id not in SerialTransport._thread_map:
            SerialTransport._thread_counter += 1
            SerialTransport._thread_map[thread_id] = SerialTransport._thread_counter
        
        thread_num = SerialTransport._thread_map[thread_id]
        entry = f"[{timestamp}] [T:{thread_num}] [{level}] {message}"
        print(entry, flush=True)

    def _generate_command_id(self) -> int:
        old_counter = self._command_counter
        self._command_counter = (self._command_counter + 1) & 0x2710
        return self._command_counter

    def find_com_port(self) -> Optional[str]:
        self._log("Starting COM port discovery")
        
        if self.override_port:
            self._log(f"Override port enabled, using: {self._fallback_com_port}")
            return self._fallback_com_port
            
        all_ports = list_ports.comports()
        self._log(f"Found {len(all_ports)} COM ports total")

        target_hwid = "VID:PID=1A86:55D3"
        
        for i, port in enumerate(all_ports):
            self._log(f"Port {i}: {port.device} - HWID: {port.hwid}")
            if target_hwid in port.hwid.upper():
                self._log(f"Target device found on port: {port.device}")
                return port.device
        
        self._log("Target device not found in COM port scan")
        
        if self._fallback_com_port:
            self._log(f"Using fallback COM port: {self._fallback_com_port}")
            return self._fallback_com_port
        
        self._log("No fallback port specified, returning None")
        return None

    def _parse_response_line(self, line: bytes) -> ParsedResponse:
        if line.startswith(b'>>> '):
            content = line[4:].decode('ascii', 'ignore').strip()
            return ParsedResponse(None, content, False)
        
        content = line.decode('ascii', 'ignore').strip()
        return ParsedResponse(None, content, False)

    
    def _handle_button_data(self, byte_val: int) -> None:
        if byte_val == self._last_button_mask:
            return

        changed_bits = byte_val ^ self._last_button_mask
        print("\n", end='')
        self._log(f"Button state changed: 0x{self._last_button_mask:02X} -> 0x{byte_val:02X}")

        for bit in range(8):
            if changed_bits & (1 << bit):
                is_pressed = bool(byte_val & (1 << bit))
                button_name = self.BUTTON_MAP[bit] if bit < len(self.BUTTON_MAP) else f"bit{bit}"
            
                self._log(f"Button {button_name}: {'PRESSED' if is_pressed else 'RELEASED'}")
                print(">>> ", end='', flush=True)
                if is_pressed:
                    self._button_states |= (1 << bit)
                else:
                    self._button_states &= ~(1 << bit)
            
                if self._button_callback and bit < len(self.BUTTON_ENUM_MAP):
                    try:
                        self._button_callback(self.BUTTON_ENUM_MAP[bit], is_pressed)
                    except Exception as e:
                        self._log(f"Button callback failed: {e}", "ERROR")

        self._last_button_mask = byte_val

    def _process_pending_commands(self, content: str) -> None:
        if not content or not self._pending_commands:
            return

        with self._command_lock:
            if not self._pending_commands:
                return

            oldest_id = next(iter(self._pending_commands))
            pending = self._pending_commands[oldest_id]

            if pending.future.done():
                return

            if content == pending.command:
                if not pending.expect_response:
                    pending.future.set_result(pending.command)
                    del self._pending_commands[oldest_id]
            else:
                pending.future.set_result(content)
                del self._pending_commands[oldest_id]

    def _cleanup_timed_out_commands(self) -> None:
        if not self._pending_commands:
            return
            
        current_time = time.time()
        
        with self._command_lock:
            timed_out = [
                (cmd_id, pending) 
                for cmd_id, pending in self._pending_commands.items()
                if current_time - pending.timestamp > pending.timeout
            ]

            for cmd_id, pending in timed_out:
                age = current_time - pending.timestamp
                self._log(f"Command '{pending.command}' timed out after {age:.3f}s", "ERROR")
                del self._pending_commands[cmd_id]
                if not pending.future.done():
                    pending.future.set_exception(
                        MakcuTimeoutError(f"Command #{cmd_id} timed out")
                    )

    def _listen(self) -> None:
        self._log("Starting listener thread")
        read_buffer = bytearray(4096)
        line_buffer = bytearray(256)
        line_pos = 0

        serial_read = self.serial.read
        serial_in_waiting = lambda: self.serial.in_waiting
        is_connected = lambda: self._is_connected
        stop_requested = self._stop_event.is_set

        last_cleanup = time.time()
        cleanup_interval = 0.05
        
        expecting_text_mode = False
        last_byte = None

        while is_connected() and not stop_requested():
            try:
                bytes_available = serial_in_waiting()
                if not bytes_available:
                    time.sleep(0.001)
                    continue
            
                bytes_read = serial_read(min(bytes_available, 4096))
                for byte_val in bytes_read:
                    
                    if last_byte == 0x0D and byte_val == 0x0A:
                        if line_pos > 0:
                            line = bytes(line_buffer[:line_pos])
                            line_pos = 0
                            if line:
                                response = self._parse_response_line(line)
                                if response.content:
                                    self._process_pending_commands(response.content)
                        expecting_text_mode = False
                    
                    elif byte_val >= 32 or byte_val in (0x09,):
                        expecting_text_mode = True
                        if line_pos < 256:
                            line_buffer[line_pos] = byte_val
                            line_pos += 1
                    
                    elif byte_val == 0x0D:
                        if expecting_text_mode or line_pos > 0:
                            expecting_text_mode = True
                        else:
                            pass
                    
                    elif byte_val == 0x0A:
                        last_byte_str = f"0x{last_byte:02X}" if last_byte is not None else "None"
                        # self._log(f"LF detected: last_byte={last_byte_str}, expecting_text={expecting_text_mode}, line_pos={line_pos}", "DEBUG")
                        
                        button_combination_detected = False
                        
                        if (self._last_button_mask != 0 or 
                            (last_byte is not None and last_byte < 32 and last_byte != 0x0D) or
                            (line_pos > 0 and not expecting_text_mode)):
                            
                            # self._log("LF: Detected as button combination (right + mouse4 = 0x0A)", "DEBUG")
                            self._handle_button_data(byte_val)
                            expecting_text_mode = False
                            button_combination_detected = True
                            line_pos = 0
                        
                        if not button_combination_detected:
                            if last_byte == 0x0D:
                                self._log("LF: Completing CRLF sequence", "DEBUG")
                                if line_pos > 0:
                                    line = bytes(line_buffer[:line_pos])
                                    line_pos = 0
                                    if line:
                                        response = self._parse_response_line(line)
                                        if response.content:
                                            self._process_pending_commands(response.content)
                                expecting_text_mode = False
                            elif line_pos > 0 and expecting_text_mode:
                                self._log("LF: Ending line with text in buffer", "DEBUG")
                                line = bytes(line_buffer[:line_pos])
                                line_pos = 0
                                if line:
                                    response = self._parse_response_line(line)
                                    if response.content:
                                        self._process_pending_commands(response.content)
                                expecting_text_mode = False
                            elif expecting_text_mode:
                                self._log("LF: Empty line in text mode", "DEBUG")
                                expecting_text_mode = False
                            else:
                                self._log("LF: Treating as button data (no text context)", "DEBUG")
                                self._handle_button_data(byte_val)
                                expecting_text_mode = False
                                line_pos = 0  # Clear any accumulated data
                    
                    elif byte_val < 32:
                        if last_byte == 0x0D:
                            self._log(f"Processing delayed CR as button data: 0x0D", "DEBUG")
                            self._handle_button_data(0x0D)
                        
                        self._handle_button_data(byte_val)
                        expecting_text_mode = False
                        line_pos = 0
                    
                    last_byte = byte_val
                            
                current_time = time.time()
                if current_time - last_cleanup > cleanup_interval:
                    self._cleanup_timed_out_commands()
                    last_cleanup = current_time
                
            except serial.SerialException as e:
                self._log(f"Serial exception in listener: {e}", "ERROR")
                if self.auto_reconnect:
                    self._attempt_reconnect()
                else:
                    break
            except Exception as e:
                self._log(f"Unexpected exception in listener: {e}", "ERROR")

        self._log("Listener thread ending")

    def _attempt_reconnect(self) -> None:
        self._log(f"Attempting reconnect #{self._reconnect_attempts + 1}/{self.MAX_RECONNECT_ATTEMPTS}")
        
        if self._reconnect_attempts >= self.MAX_RECONNECT_ATTEMPTS:
            self._log("Max reconnect attempts reached, giving up", "ERROR")
            self._is_connected = False
            return
        
        self._reconnect_attempts += 1
        
        try:
            if self.serial and self.serial.is_open:
                self._log("Closing existing serial connection for reconnect")
                self.serial.close()
            
            time.sleep(self.RECONNECT_DELAY)
            
            self.port = self.find_com_port()
            if not self.port:
                raise MakcuConnectionError("Device not found during reconnect")
            
            self._log(f"Reconnecting to {self.port} at {self.baudrate} baud")
            self.serial = serial.Serial(
                self.port, 
                self.baudrate, 
                timeout=0.001,
                write_timeout=0.01
            )
            
            if not self._change_baud_to_4M():
                raise MakcuConnectionError("Failed to change baud during reconnect")
            
            if self.send_init:
                self._log("Sending init command during reconnect")
                self.serial.write(b"km.buttons(1)\r")
                self.serial.flush()
            
            self._reconnect_attempts = 0
            self._log("Reconnect successful")
            
        except Exception as e:
            self._log(f"Reconnect attempt failed: {e}", "ERROR")
            time.sleep(self.RECONNECT_DELAY)

    def _change_baud_to_4M(self) -> bool:
        self._log("Changing baud rate to 4M")
        
        if self.serial and self.serial.is_open:
            self.serial.write(self.BAUD_CHANGE_COMMAND)
            self.serial.flush()
            
            time.sleep(0.02)
            
            old_baud = self.serial.baudrate
            self.serial.baudrate = 4000000
            self._current_baud = 4000000
            
            self._log(f"Baud rate changed: {old_baud} -> {self.serial.baudrate}")
            return True
        
        self._log("Cannot change baud - serial not open", "ERROR")
        return False

    def connect(self) -> None:
        connection_start = time.time()
        self._log("Starting connection process")
        
        if self._is_connected:
            self._log("Already connected")
            return
        
        if not self.override_port:
            self.port = self.find_com_port()
        else:
            self.port = self._fallback_com_port
            
        if not self.port:
            raise MakcuConnectionError("Makcu device not found")
        
        self._log(f"Connecting to {self.port}")
        
        try:
            self.serial = serial.Serial(
                self.port, 
                115200, 
                timeout=0.001,
                write_timeout=0.01,
                xonxoff=False,
                rtscts=False,
                dsrdtr=False
            )
            
            if not self._change_baud_to_4M():
                raise MakcuConnectionError("Failed to switch to 4M baud")
            
            self._is_connected = True
            self._reconnect_attempts = 0
            
            connection_time = time.time() - connection_start
            self._log(f"Connection established in {connection_time:.3f}s")
            
            if self.send_init:
                self._log("Sending initialization command")
                init_cmd = b"km.buttons(1)\r"
                self.serial.write(init_cmd)
                self.serial.flush()

            self._stop_event.clear()
            self._listener_thread = threading.Thread(
                target=self._listen, 
                daemon=True,
                name="MakcuListener"
            )
            self._listener_thread.start()
            self._log(f"Listener thread started: {self._listener_thread.name}")
            
        except Exception as e:
            self._log(f"Connection failed: {e}", "ERROR")
            if self.serial:
                try:
                    self.serial.close()
                except:
                    pass
            raise MakcuConnectionError(f"Failed to connect: {e}")

    def disconnect(self) -> None:
        self._log("Starting disconnection process")
        
        self._is_connected = False
        
        if self.send_init:
            self._stop_event.set()
            if self._listener_thread and self._listener_thread.is_alive():
                self._listener_thread.join(timeout=0.1)
                if self._listener_thread.is_alive():
                    self._log("Listener thread did not join within timeout")
                else:
                    self._log("Listener thread stopped")
        
        pending_count = len(self._pending_commands)
        if pending_count > 0:
            self._log(f"Cancelling {pending_count} pending commands")
        
        with self._command_lock:
            for cmd_id, pending in self._pending_commands.items():
                if not pending.future.done():
                    pending.future.cancel()
            self._pending_commands.clear()
        
        if self.serial and self.serial.is_open:
            self._log(f"Closing serial port: {self.serial.port}")
            self.serial.close()
            
        self.serial = None
        self._log("Disconnection completed")

    def send_command(self, command: str, expect_response: bool = False, 
                timeout: float = DEFAULT_TIMEOUT) -> Optional[str]:
        command_start = time.time()
        
        if not self._is_connected or not self.serial or not self.serial.is_open:
            raise MakcuConnectionError("Not connected")
        
        if not expect_response:
            self.serial.write(f"{command}\r\n".encode('ascii'))
            self.serial.flush()
            send_time = time.time() - command_start
            self._log(f"Command '{command}' sent in {send_time:.5f}s (no response expected)")
            return command
        
        cmd_id = self._generate_command_id()
        tagged_command = f"{command}#{cmd_id}"
        
        future = Future()
        
        with self._command_lock:
            self._pending_commands[cmd_id] = PendingCommand(
                command_id=cmd_id,
                command=command,
                future=future,
                timestamp=time.time(),
                expect_response=expect_response,
                timeout=timeout
            )
        
        try:
            cmd_bytes = f"{tagged_command}\r\n".encode('ascii')
            self.serial.write(cmd_bytes)
            self.serial.flush()
            
            result = future.result(timeout=timeout)
            
            response = result.split('#')[0] if '#' in result else result
            total_time = time.time() - command_start
            self._log(f"Command '{command}' completed in {total_time:.5f}s total")
            return response
            
        except TimeoutError:
            total_time = time.time() - command_start
            self._log(f"Command '{command}' timed out after {total_time:.3f}s", "ERROR")
            raise MakcuTimeoutError(f"Command timed out: {command}")
        except Exception as e:
            total_time = time.time() - command_start
            self._log(f"Command '{command}' failed after {total_time:.3f}s: {e}", "ERROR")
            with self._command_lock:
                self._pending_commands.pop(cmd_id, None)
            raise

    async def async_send_command(self, command: str, expect_response: bool = False,
                               timeout: float = DEFAULT_TIMEOUT) -> Optional[str]:
        self._log(f"Async sending command: '{command}'")
        loop = asyncio.get_running_loop()
        return await loop.run_in_executor(
            None, self.send_command, command, expect_response, timeout
        )

    def is_connected(self) -> bool:
        connected = self._is_connected and self.serial is not None and self.serial.is_open
        return connected

    def set_button_callback(self, callback: Optional[Callable[[MouseButton, bool], None]]) -> None:
        self._log(f"Setting button callback: {callback is not None}")
        self._button_callback = callback

    def get_button_states(self) -> Dict[str, bool]:
        states = {
            self.BUTTON_MAP[i]: bool(self._button_states & (1 << i))
            for i in range(5)
        }
        return states

    def get_button_mask(self) -> int:
        return self._last_button_mask

    def enable_button_monitoring(self, enable: bool = True) -> None:
        cmd = "km.buttons(1)" if enable else "km.buttons(0)"
        self._log(f"{'Enabling' if enable else 'Disabling'} button monitoring")
        self.send_command(cmd)

    async def __aenter__(self):
        self._log("Async context manager enter")
        loop = asyncio.get_running_loop()
        await loop.run_in_executor(None, self.connect)
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        self._log("Async context manager exit")
        loop = asyncio.get_running_loop()
        await loop.run_in_executor(None, self.disconnect)

    def __enter__(self):
        self._log("Sync context manager enter")
        self.connect()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._log("Sync context manager exit")
        self.disconnect()


from typing import List
from .controller import MakcuController

__version__: str
__all__: List[str]

def create_controller(
    fallback_com_port: str = "", 
    debug: bool = False, 
    send_init: bool = True
) -> MakcuController: ...